@@ -1,7 +1,7 @@
 #
 #  BeanCounter.pm --- A stock portfolio performance monitoring toolkit
 #
-#  Copyright (C) 1998 - 2005  Dirk Eddelbuettel <edd@debian.org>
+#  Copyright (C) 1998 - 2006  Dirk Eddelbuettel <edd@debian.org>
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
 #  along with this program; if not, write to the Free Software
 #  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-#  $Id: BeanCounter.pm,v 1.85 2005/06/11 00:46:00 edd Exp $
+#  $Id: BeanCounter.pm,v 1.103 2007/10/04 03:23:25 edd Exp $
 
 package Finance::BeanCounter;
 
@@ -42,6 +42,7 @@ use Text::ParseWords;		# parse .csv data more reliably
 	     DatabaseDailyData
 	     DatabaseHistoricalData
 	     DatabaseHistoricalFXData
+	     DatabaseHistoricalUBCFX
 	     DatabaseInfoData
 	     ExistsDailyData
 	     ExistsFXDailyData
@@ -51,9 +52,11 @@ use Text::ParseWords;		# parse .csv data more reliably
 	     GetDate
 	     GetDailyData
 	     GetFXData
+	     GetFXDatum
 	     GetUBCFXData
 	     GetUBCFXHash
-	     GetFXMaps
+	     GetYahooCurrency
+	     GetIsoCurrency
 	     GetHistoricalData
 	     GetPortfolioData
 	     GetPriceData
@@ -66,13 +69,13 @@ use Text::ParseWords;		# parse .csv data more reliably
 	     Sign
 	     UpdateDatabase
 	     UpdateFXDatabase
-	     UpdateFXviaPACIFIC 
+	     UpdateFXviaUBC 
 	     UpdateTimestamp
 	    );
 @EXPORT_OK = qw( );
 %EXPORT_TAGS = (all => [@EXPORT_OK]);
 
-my $VERSION = sprintf("%d.%d", q$Revision: 1.85 $ =~ /(\d+)\.(\d+)/); 
+my $VERSION = sprintf("%d.%d", q$Revision: 1.103 $ =~ /(\d+)\.(\d+)/); 
 
 my %Config;			# local copy of configuration hash
 
@@ -83,39 +86,32 @@ sub BeanCounterVersion {
 
 
 sub ConnectToDb {		# log us into the database (PostgreSQL)
-  my $dbh = undef;
-
+  my $hoststr = '';
+  $hoststr = "host=$Config{host}"
+	unless (grep(/^$Config{host}$/, ('localhost','127.0.0.1','::1/128')));
+  my $dsn = 'dbi:';
   if ($Config{odbc}) {
-    $dbh = DBI->connect("dbi:ODBC:$Config{dsn}",
-			$Config{user}, $Config{passwd}, 
-			{ PrintError => $Config{debug}, 
-			  Warn => $Config{verbose}, 
-			  AutoCommit => 0 });
+      $dsn .= "ODBC:$Config{dsn}";
   } elsif (lc $Config{dbsystem} eq "postgresql") {
-    $dbh = DBI->connect("dbi:Pg:dbname=$Config{dbname};host=$Config{host}",
-			$Config{user}, $Config{passwd},
-			{ PrintError => $Config{debug}, 
-			  Warn => $Config{verbose}, 
-			  AutoCommit => 0 });
+      $dsn .= "Pg:dbname=$Config{dbname};${hoststr}";
   } elsif (lc $Config{dbsystem} eq "mysql") {
-    $dbh = DBI->connect("dbi:mysql:dbname=$Config{dbname};host=$Config{host}",
-			$Config{user}, $Config{passwd},
-			{ PrintError => $Config{debug}, 
-			  Warn => $Config{verbose}, 
-			  AutoCommit => 0 });
+      $dsn .= "mysql:dbname=$Config{dbname};${hoststr}";
   } elsif (lc $Config{dbsystem} eq "sqlite") {
-    $dbh = DBI->connect("dbi:SQLite:dbname=$Config{dbname}","","",
-			{ PrintError => $Config{debug}, 
-			  Warn => $Config{verbose}, 
-			  AutoCommit => 0 });
+      $dsn .= "SQLite:dbname=$Config{dbname}";
+      $Config{user} = '';
+      $Config{passwd} = '';
   } elsif (lc $Config{dbsystem} eq "sqlite2") {
-    $dbh = DBI->connect("dbi:SQLite2:dbname=$Config{dbname}","","",
-			{ PrintError => $Config{debug}, 
-			  Warn => $Config{verbose}, 
-			  AutoCommit => 0 });
+      $dsn .= "SQLite2:dbname=$Config{dbname}";
+      $Config{user} = '';
+      $Config{passwd} = '';
   } else {
     die "Database system $Config{dbsystem} is not supported\n";
   }
+  my $dbh = DBI->connect($dsn, $Config{user}, $Config{passwd}, 
+			 { PrintError => $Config{debug}, 
+			   Warn => $Config{verbose}, 
+			   AutoCommit => 0 });
+  
   die "No luck with database connection\n" unless ($dbh);
 
   return $dbh;
@@ -124,7 +120,7 @@ sub ConnectToDb {		# log us into the database (PostgreSQL)
 
 sub CloseDB {
   my $dbh = shift;
-  $dbh->disconnect;
+  $dbh->disconnect or warn $dbh->errstr;
 }
 
 
@@ -183,7 +179,7 @@ sub GetTodaysAndPreviousDates {
 sub GetConfig {
   my ($file, $debug, $verbose, $fx, $extrafx, $updatedate,
       $dbsystem, $dbname, $fxupdate, $commit, $equityupdate, 
-      $pacificfx, $command) = @_;
+      $ubcfx, $hostarg, $command) = @_;
 
   %Config = ();			# reset hash
 
@@ -206,9 +202,6 @@ sub GetConfig {
   # DSN name for ODBC
   $Config{dsn} = "beancounter";	# default ODBC data source name
 
-  # host is needed only for the DBI-Pg or DBI-mysql interface
-  $Config{host} = "localhost";	# default to local machine
-
   # default to updating FX
   if ($fxupdate) {
     $Config{fxupdate} = 1;
@@ -231,11 +224,13 @@ sub GetConfig {
   }
 
   # default to updateing stocks too
-  if ($pacificfx) {
-    $Config{pacificfx} = 1;
+  if ($ubcfx) {
+    $Config{ubcfx} = 1;
   } else {
-    $Config{pacificfx} = 0;
+    $Config{ubcfx} = 0;
   }
+  # pre-load a default host argument
+  $Config{host} = $hostarg if defined($hostarg);
 
   unless ( -f $file ) {
     warn "Config file $file not found, ignored.\n";
@@ -257,6 +252,11 @@ sub GetConfig {
   $Config{dbsystem} = $dbsystem if defined($dbsystem);
   $Config{odbc} = 1 if defined($dbsystem) and lc $dbsystem eq "odbc";
 
+  # but allow command-line argument to override 
+  $Config{host} = $hostarg 	
+      if defined($hostarg) and $hostarg ne "localhost";	
+
+
   if (defined($extrafx)) {
     unless ($command =~ /^(update|dailyjob)$/) {
       warn "Warning: --extrafx ignored as not updating db\n";
@@ -301,6 +301,7 @@ sub GetCashData {
     $cash{$name}{fx} = $fx;
     $cash{$name}{cost} = $cost;
   }
+  $sth->finish();
   return(\%cash);
 }
 
@@ -401,29 +402,31 @@ sub GetUBCFXHash {
   return \%res;			# return the new hash
 }
 
-# map between ISO country codes and Yahoo symbols for the Philly exchange
-sub GetFXMaps {
-  my %iso2yahoo = (
-		   "AUD" => "^XAD",  	# was "AUDUSD=X",
-		   "CAD" => "^XCD", 	# was "CADUSD=X",
-		   "CHF" => "^XSF",	# was "CHFUSD=X",
-		   "EUR" => "^XEU",	# was "EURUSD=X",
-		   "GBP" => "^XBP",     # was "GBPUSD=X",
-		   "JPY" => "^XJY",	# was "JPYUSD=X",
-		   "USD" => "----"
-		   ## deprecated  "DEM" => "^XDM"
-		  );
-  my %yahoo2iso = (
-		   "^XAD" => "AUD",
-		   "^XCD" => "CAD",
-		   "^XSF" => "CHF",
-		   "^XEU" => "EUR",
-		   "^XBP" => "GBP",
-		   "^XJY" => "JPY",
-		   "----" => "USD"
-		   ## deprecated  "^XDM" => "DEM"
-		  );
-  return (\%iso2yahoo, \%yahoo2iso);
+sub getIso2YahooCurrencyHashRef() {
+    # map between ISO country codes and Yahoo symbols for the Philly exchange
+    return {"AUD" => "^XAY", # was "^XAD", "AUDUSD=X",
+	    "CAD" => "^XCV", # was "^XCD", "CADUSD=X",
+	    "CHF" => "^XSY", # was "^XSF", "CHFUSD=X",
+	    "EUR" => "^XEU", # was "EURUSD=X",
+	    "GBP" => "^XBX", # was "^XBP", "GBPUSD=X",
+	    "JPY" => "^XJZ", # was "^XJY", "JPYUSD=X",
+	    "USD" => "----"};
+}
+
+
+sub GetYahooCurrency($) {
+    my ($isoCurrency) = @_;
+    my $ref = getIso2YahooCurrencyHashRef();
+    return $ref->{$isoCurrency};
+}
+
+
+sub GetIsoCurrency($) {
+    my ($yahooCurrency) = @_;
+    my $ref = getIso2YahooCurrencyHashRef();
+    # Reverse the hash table, ie. yahoo => iso:
+    my %yahoo2isoHash = map { $ref->{$_} => $_ } keys(%$ref);
+    return $yahoo2isoHash{$yahooCurrency};
 }
 
 
@@ -484,7 +487,7 @@ sub GetPriceData {
   @symbols = @{ $dbh->selectcol_arrayref($stmt) };
 
   # for each symbol, get most recent date subject to supplied date
-  $stmt  = qq{select max(date) 
+  $stmt  = qq{select max(date)
 	      from stockprices 
 	      where symbol = ? 
 	      and day_close > 0
@@ -493,17 +496,23 @@ sub GetPriceData {
   print "GetPriceData():\n\$stmt = $stmt\n" if $Config{debug};
 
   # for each symbol, get most recent date subject to supplied date:\n";
-  $sth = $dbh->prepare($stmt);
   foreach $ra (@symbols) {	
+    if (!defined($sth)) {
+      $sth = $dbh->prepare($stmt);
+    }
     $rv = $sth->execute($ra, $date); # run query for report end date
     my $res = $sth->fetchrow_array;
     $dates{$ra} = $res;
     $sth->finish() if $Config{odbc};
   }
 
+#sum(p.shares*p.cost)/sum(p.shares) as p.cost, 
   # now get closing price etc at date
   $stmt =    qq{select i.symbol, i.name, p.shares, p.currency,
-		       d.day_close, p.cost, p.date, d.previous_close
+		       d.day_close, 
+		       p.cost, 
+		       p.date, 
+		       d.previous_close
 		from stockinfo i, portfolio p, stockprices d
 		where d.symbol = p.symbol
 		and i.symbol = d.symbol
@@ -539,12 +548,18 @@ sub GetPriceData {
   $stmt .= qq{ and d.symbol in
 	      (select distinct symbol from portfolio where $res)
 	     }   if (defined($res));
+##  $stmt .= qq{ group by 	      i.symbol,i.name,p.shares,p.currency,d.day_close,p.date,d.previous_close };
+
+#select symbol, avg('today'-date) as days, sum(shares*cost)/sum(shares) as cost, sum(shares) as size, sum(shares*cost) as pos from portfolio where owner!='peter' group by symbol order by days desc;
   print "GetPriceData():\n\$stmt = $stmt\n" if $Config{debug};
 
   # now get closing price etc at date
-  $sth = $dbh->prepare($stmt);
+  $sth = undef;
   my $i = 0;
   foreach $ra (@symbols) {		
+    if (!defined($sth)) {
+      $sth = $dbh->prepare($stmt);
+    }
     $rv = $sth->execute($dates{$ra}, $ra);
     while (($symbol, $name, $shares, $currency, $price, 
 	    $cost, $pdate, $prevprice) = $sth->fetchrow_array) {
@@ -561,7 +576,10 @@ sub GetPriceData {
       $name .= ":$i";
       $i++;
       $shares{$name} = $shares;
+      $purchdate{$name} = $pdate; # also store purchuse date on non-aggregate entry
+      $cost{$name} = $cost;	  # also store purchuse cost on non-aggregate entry
     }
+    $sth->finish;
   }
 
   print Dumper(\%prices) if $Config{debug};
@@ -587,9 +605,12 @@ sub GetFXData {
 
   # get most recent date subject to supplied date
   my %fxdates;
-  my $sth = $dbh->prepare($stmt);
+  my $sth;
   foreach my $fxval (sort values %$fx) {
     next if $fxval eq 'USD';	# skip USD
+    if (!defined($sth)) {
+      $sth = $dbh->prepare($stmt);
+    }
     $rv = $sth->execute($fxval, $date); # run query for report end date
     my $res = $sth->fetchrow_array;
     $fxdates{$fxval} = $res;
@@ -602,13 +623,16 @@ sub GetFXData {
 	    };
   print "GetFXData():\n\$stmt = $stmt\n" if $Config{debug};
 
-  $sth = $dbh->prepare($stmt);
+  $sth = undef;
   my (%fx_prices,%prev_fx_prices);
   foreach my $fxval (sort values %$fx) {
     if ($fxval eq 'USD') {	
       $fx_prices{$fxval} = 1.0;
       $prev_fx_prices{$fxval} = 1.0;
     } else {
+      if (!defined($sth)) {
+        $sth = $dbh->prepare($stmt);
+      }
       $sth->execute($fxdates{$fxval}, $fxval);	# run query for FX cross
       my ($val, $prevval) = $sth->fetchrow_array
 	or die "Found no $fxval for $date in the beancounter database.\n " .
@@ -625,6 +649,15 @@ sub GetFXData {
   return (\%fx_prices, \%prev_fx_prices);
 }
 
+## simple wrapper for GetFXDate for single currency + date
+sub GetFXDatum {		
+  my ($dbh, $date, $fx) = @_;
+
+  my %fxhash; 
+  $fxhash{foo} = $fx;
+  my ($fxcurrent) = GetFXData($dbh, $date, \%fxhash); 
+  return $fxcurrent->{$fx};
+}
 
 ## NB no longer used as we employ Finance::YahooQuote directly
 sub GetQuote {			# taken from Dj's Finance::YahooQuote
@@ -928,9 +961,14 @@ sub GetRiskData {
   # eg GNU Octave), we flatten the computation into a double loop
   my $sum = 0;
   foreach my $pkey (keys %pos) {
-    foreach my $vkey (keys %vol) { 
-      $sum += $pos{$pkey} * $pos{$vkey} * $vol{$vkey} * $vol{$pkey} 
-	* $cor{$vkey}{$pkey};
+    if (defined($pos{$pkey}) && defined($vol{$pkey})) {
+      foreach my $vkey (keys %vol) { 
+	if (defined($pos{$vkey}) && defined($vol{$vkey}) &&
+	    defined($cor{$vkey}{$pkey})) {
+	  $sum += $pos{$pkey} * $pos{$vkey} * $vol{$vkey} * $vol{$pkey} * 
+	    $cor{$vkey}{$pkey};
+        }
+      }
     }
   }
   my $var = $crit * sqrt($sum);
@@ -943,9 +981,14 @@ sub GetRiskData {
     my $sum = 0;
     $pos{$outer} = 0;
     foreach my $pkey (keys %pos) {
-      foreach my $vkey (keys %vol) { 
-	$sum += $pos{$pkey} * $pos{$vkey} * $vol{$vkey} * $vol{$pkey} 
-	  * $cor{$vkey}{$pkey};
+      if (defined($pos{$pkey}) && defined($vol{$pkey})) {
+        foreach my $vkey (keys %vol) { 
+	  if (defined($pos{$vkey}) && defined($vol{$vkey}) &&
+	      defined($cor{$vkey}{$pkey})) {
+            $sum += $pos{$pkey} * $pos{$vkey} * $vol{$vkey} * $vol{$pkey} 
+	            * $cor{$vkey}{$pkey};
+	  }
+	}
       }
     }
     $margvar{$outer} = $crit * sqrt($sum) - $var;
@@ -958,7 +1001,27 @@ sub GetRiskData {
 
 sub DatabaseDailyData {		# a row to the dailydata table
   my ($dbh, %hash) = @_;
-  my $cmd;
+  my @cols = ('previous_close', 'day_open', 'day_high', 'day_low',
+	      'day_close', 'day_change', 'bid', 'ask', 'volume');
+  my @updTerms = ();
+  foreach my $col (@cols) {
+    push(@updTerms, "$col = ?");
+  }
+  my $updStmt = 'update stockprices set ' . join(', ', @updTerms) .
+      ' where symbol = ? and date = ?';
+  print "$updStmt\n" if $Config{debug};
+  my $updSth;
+
+  push(@cols, 'symbol', 'date');
+  my @insTerms = ();
+  foreach my $col (@cols) {
+    push(@insTerms, '?');
+  }
+  my $insStmt = 'insert into stockprices (' . join(', ', @cols) .
+      ') values (' . join(', ', @insTerms) . ')';
+  print "$insStmt\n" if $Config{debug};
+  my $insSth;
+  
   foreach my $key (keys %hash) { # now split these into reference to the arrays
     print "$hash{$key}{symbol} " if $Config{verbose};
 
@@ -968,57 +1031,56 @@ sub DatabaseDailyData {		# a row to the dailydata table
     }
 
     if (ExistsDailyData($dbh, %{$hash{$key}})) {
-      $cmd = "update stockprices set";
-      $cmd.= " previous_close = $hash{$key}{previous_close}," 
-	if ($hash{$key}{previous_close} !~ /'?N\/A'?/ );
-      $cmd.= " day_open       = $hash{$key}{day_open}, " 
-	if ($hash{$key}{day_open} !~ /'?N\/A'?/ );
-      $cmd.= " day_low        = $hash{$key}{day_low}, " 
-	if ($hash{$key}{day_low} !~ /'?N\/A'?/ );
-      $cmd.= " day_high       = $hash{$key}{day_high}, " 
-	if ($hash{$key}{day_high} !~ /'?N\/A'?/ );
-      $cmd.= " day_close      = $hash{$key}{day_close}, ";
-      $cmd.= " day_change     = $hash{$key}{day_change}, " 
-	if ($hash{$key}{day_change} !~ /'?N\/A'?/ );
-      $cmd.= " bid            = $hash{$key}{bid}, " 
-	if ($hash{$key}{bid} !~ /'?N\/A'?/ );
-      $cmd.= " ask            = $hash{$key}{ask}, " 
-	if ($hash{$key}{ask} !~ /'?N\/A'?/ );
-      $cmd.= " volume         = $hash{$key}{volume}, " 
-	if ($hash{$key}{volume} ne '0' );
-      $cmd.= "where symbol     = '$hash{$key}{symbol}' ";
-      $cmd.= "and date         = '$hash{$key}{date}'";
-      $cmd =~ s/, where/ where/;
-    } else {
-      $cmd = "insert into stockprices values (" . 
-             "'$hash{$key}{symbol}'," . 
-	     "'$hash{$key}{date}'," .
-	     "$hash{$key}{previous_close}," .
-	     "$hash{$key}{day_open}," .
-	     "$hash{$key}{day_low}," .
-	     "$hash{$key}{day_high}," .
-	     "$hash{$key}{day_close}," .
-	     "$hash{$key}{day_change}," .
-	     "$hash{$key}{bid}," .
-	     "$hash{$key}{ask}," .
-	     "$hash{$key}{volume})";
+      my @vals = ();
+      foreach my $col (@cols) {
+	  if ($hash{$key}{$col} =~ m/^\s*N\/A\s*$/) {
+	      push(@vals, undef);
+	  } else {
+	      push(@vals, $hash{$key}{$col});
+	  }
+      }
+      if ($Config{commit}) {
+          if (!defined($updSth)) {
+              $updSth = $dbh->prepare($updStmt) or die $dbh->errstr;
+          }
+          $updSth->execute(@vals)
+              and $updSth->finish()
+              or warn $dbh->errstr . "Update failed for " .
+	      	"$hash{$key}{symbol} with [$updStmt]\n";
+      }
     }
-    $cmd =~ s|'?N/A'?|null|g;	# convert (textual) "N/A" into (database) null 
-    print "$cmd\n" if $Config{debug};
-    print "$hash{$key}{symbol} " if $Config{verbose};
-    if ($Config{commit}) {
-      $dbh->do($cmd) or warn "Failed for $hash{$key}{symbol} with $cmd\n";
-      $dbh->commit();
+    else {
+      my @vals = ();
+      foreach my $col (@cols) {
+	  if ($hash{$key}{$col} =~ m/^\s*N\/A\s*$/) {
+	      push(@vals, undef);
+	  } else {
+	      push(@vals, $hash{$key}{$col});
+	  }
+      }
+      if ($Config{commit}) {
+          if (!defined($insSth)) {
+              $insSth = $dbh->prepare($insStmt) or die $dbh->errstr;
+          }
+          $insSth->execute(@vals)
+              and $insSth->finish()
+              or warn $dbh->errstr . "Insert failed for " .
+	      	"$hash{$key}{symbol} with [$insStmt]\n";
+      }
     }
   }
+  $dbh->commit() if $Config{commit};
 }
 
 
 sub DatabaseFXDailyData {
   my ($dbh, %hash) = @_;
-  my ($iso2yahoo,$yahoo2iso) = GetFXMaps;
   foreach my $key (keys %hash) { # now split these into reference to the arrays
-    my $fx = $yahoo2iso->{$hash{$key}{symbol}};
+    if ($key eq "") {
+ 	print "Empty key in DatabaseFXDailyData, skipping\n" if $Config{debug};
+	next;
+    }
+    my $fx = GetIsoCurrency($hash{$key}{symbol});
     print "$fx ($hash{$key}{symbol})  " if $Config{debug};
     if (ExistsFXDailyData($dbh, $fx, %{$hash{$key}})) {
       # different sequence of parameters, see SQL statement above!
@@ -1120,99 +1182,86 @@ sub DatabaseFXDailyData {
 
 sub DatabaseHistoricalData {
   my ($dbh, $symbol, @res) = @_;
-  my $checked = 0;		# flag to ensure not nonsensical or errors
-  my %data;			# hash to store data of various completenesses
   $symbol = uc $symbol;		# make sure symbols are uppercase'd
-  foreach $ARG (@res) {		# loop over all supplied symbols
-    next if m/^<\!-- .*-->/;    # skip lines with html comments (April 2004)
-    # make sure the first line of data is correct so we don't insert garbage
-    if ($checked==0 and m/Date(,Open,High,Low)?,Close(,Volume)?/) {
-      $checked = tr/,//;
-      print "Checked now $checked\n" if $Config{verbose};
-    } elsif ($checked) {
-      my ($date, $open, $high, $low, $close, $volume, $cmd, $adjclose);
-      # based on the number of elements, ie columns, we split the parsing
-      if ($checked eq 6) {	# checked eq 6, so process adj. close
-	($date, $open, $high, $low, $close, $volume, $adjclose) = 
-	  split(/\,/, $ARG);
-	$date = UnixDate(ParseDate($date), "%Y%m%d");
-        if ( $adjclose != $close && $close != 0 ) { # process split adjustment factor
-          my $split_adj = ($adjclose / $close);
-	  %data = (symbol    => $symbol,
-		   date      => $date,
-		   day_open  => $open * $split_adj,
-		   day_high  => $high * $split_adj,
-		   day_low   => $low  * $split_adj,
-		   day_close => $adjclose,
-		   volume    => $volume);
-        } else {		# no split adjustment necessary
-          %data = (symbol    => $symbol,
-                   date      => $date,
-                   day_open  => $open,
-                   day_high  => $high,
-                   day_low   => $low,
-                   day_close => $close,
-                   volume    => $volume);
-	}
-      } elsif ($checked eq 5) { # $checked eq 5
-	($date, $open, $high, $low, $close, $volume) = split(/\,/, $ARG);
-	$date = UnixDate(ParseDate($date), "%Y%m%d");
-	%data = (symbol    => $symbol,
-		 date	   => $date,
-		 day_open  => $open,
-		 day_high  => $high,
-		 day_low   => $low,
-		 day_close => $close,
-		 volume    => $volume);
-      } elsif ($checked eq 1) {	# only close for mutual funds
-	($date, $close) = split(/\,/, $ARG);
-	$date = UnixDate(ParseDate($date), "%Y%m%d");
-	%data = (symbol    => $symbol,
-		 date	   => $date,
-		 day_close => $close);
-      } else {			# no volume for indices
-	($date, $open, $high, $low, $close) = split(/\,/, $ARG);
-	$date = UnixDate(ParseDate($date), "%Y%m%d");
-	%data = (symbol    => $symbol,
-		 date	   => $date,
-		 day_open  => $open,
-		 day_high  => $high,
-		 day_low   => $low,
-		 day_close => $close);
+
+  my %data = (symbol    => $symbol,
+	      date      => undef,
+	      day_open  => undef, 
+	      day_high  => undef,
+	      day_low   => undef, 
+	      day_close => undef,
+	      volume    => undef);
+
+  my @colNames = sort(keys(%data));
+  my @colRepl = ();
+  my @updTerms = ();
+  foreach my $col (@colNames) {
+      push(@colRepl, '?');
+      next if ($col eq 'symbol' || $col eq 'date');
+      push(@updTerms, "$col = ?");
+  }
+
+  my $insStmt = 'insert into stockprices (' . join(', ', @colNames) .
+      ') values (' . join(', ', @colRepl) . ')';
+  my $insSth;
+  my $updStmt = 'update stockprices set ' . join(', ', @updTerms) .
+      ' where symbol = ? and date = ?';
+  my $updSth;
+  print "DatabaseHistoricalData: insStmt is \"$insStmt\"\n" if $Config{debug};
+  print "DatabaseHistoricalData: updStmt is \"$updStmt\"\n" if $Config{debug};
+  
+  foreach my $line (@res) {		# loop over all supplied symbols
+    next if !defined($line);
+    ($data{date}, $data{day_open}, $data{day_high},
+     $data{day_low}, $data{day_close}, $data{volume},
+     $data{adjclose}) = split(/\,/, $line);
+    $data{date} = GetDate($data{date});
+    if (defined($data{date})) {
+      # If close was not supplied, we assume a mutual fund.
+      # So let close be open.
+      if (!defined($data{day_close})) {
+	$data{day_close} = $data{day_open};
+	$data{day_open} = undef;
+      }
+      elsif (defined($data{adjclose}) &&
+	     $data{adjclose} != $data{day_close} &&
+	     $data{day_close} != 0) { # process split adjustment factor
+	my $split_adj = $data{adjclose} / $data{day_close};
+	$data{day_open} *= $split_adj;
+	$data{day_high} *= $split_adj;
+	$data{day_low}  *= $split_adj;
+	$data{day_close} = $data{adjclose};
       }
 
-      # now given the data, decide whether we add new data or update old data
-      if (ExistsDailyData($dbh,%data)) { # update data if it exists
-	$cmd = "update stockprices set ";
-	$cmd .= "volume    = $data{volume},"    if defined($data{volume});
-	$cmd .= "day_open  = $data{day_open},"  if defined($data{day_open});
-	$cmd .= "day_low   = $data{day_low},"   if defined($data{day_low});
-	$cmd .= "day_high  = $data{day_high},"  if defined($data{day_high});
-	$cmd .= "day_close = $data{day_close} "   .
-	        "where symbol = '$data{symbol}' " .
-		"and date     = '$data{date}'";
-      } else {			# insert
-	$cmd = "insert into stockprices (symbol, date,";
-	$cmd .= "day_open," if defined($data{day_open});
-	$cmd .= "day_high," if defined($data{day_high});
-	$cmd .= "day_low,"  if defined($data{day_low});
-	$cmd .= "day_close";
-	$cmd .= ",volume"  if defined($data{volume});
-	$cmd .= ") values ('$symbol', '$date', ";
-	$cmd .= "$data{day_open},"   if defined($data{day_open});
-	$cmd .= "$data{day_high},"   if defined($data{day_high});
-	$cmd .= "$data{day_low},"    if defined($data{day_low});
-	$cmd .= "$data{day_close}"; 
-	$cmd .= ",$data{volume} "    if defined($data{volume});
-        $cmd .= ");";
+      if (ExistsDailyData($dbh, %data)) {
+	my @colVals = ();
+	foreach my $col (@colNames) {
+	  next if ($col eq 'symbol' || $col eq 'date');
+	  $data{$col} = 'NULL' if !defined($data{$col});
+	  push(@colVals, $data{$col});
+        }
+	push(@colVals, $data{symbol}, $data{date});
+	if (!defined($updSth)) {
+	    $updSth = $dbh->prepare($updStmt) or die $dbh->errstr;
+	}
+	$updSth->execute(@colVals) or die $updSth->errstr;
+	$updSth->finish();
+      }
+      else {
+	my @colVals = ();
+	foreach my $col (@colNames) {
+	  $data{$col} = 'NULL' if !defined($data{$col});
+	  push(@colVals, $data{$col});
+        }
+	if (!defined($insSth)) {
+	    $insSth = $dbh->prepare($insStmt) or die $dbh->errstr;
+	}
+	$insSth->execute(@colVals) or die $insSth->errstr;
+	$insSth->finish();
       }
-      print "$cmd\n" if $Config{debug};
-      $dbh->do($cmd) or die $dbh->errstr;
-      $dbh->commit();
-    } else {
-      ;				# do nothing with bad data
     }
   }
+  $dbh->commit() if $Config{commit};
   print "Done with $symbol\n" if $Config{verbose};
 }
 
@@ -1224,8 +1273,7 @@ sub DatabaseHistoricalFXData {
 
   my $cut = UnixDate(ParseDate("30-Dec-2003"), "%Y%m%d");
 
-  my ($iso2yahoo,$yahoo2iso) = GetFXMaps;
-  my $fx = $yahoo2iso->{$symbol};
+  my $fx = GetIsoCurrency($symbol);
   foreach $ARG (@res) {		# loop over all supplied symbols
     next if m/^<\!-- .*-->/;    # skip lines with html comments (April 2004)
     # make sure the first line of data is correct so we don't insert garbage
@@ -1281,9 +1329,11 @@ sub DatabaseHistoricalFXData {
 	##$cmd .= ",$data{volume} "    if defined($data{volume});
         $cmd .= ");";
       }
-      print "$cmd\n" if $Config{debug};
-      $dbh->do($cmd) or die $dbh->errstr;
-      $dbh->commit();
+      if ($Config{commit}) {
+	print "$cmd\n" if $Config{debug};
+	$dbh->do($cmd) or die $dbh->errstr;
+	$dbh->commit();
+      }
     } else {
       ;				# do nothing with bad data
     }
@@ -1291,6 +1341,36 @@ sub DatabaseHistoricalFXData {
   print "Done with $fx (using $symbol)\n" if $Config{verbose};
 }
 
+sub DatabaseHistoricalUBCFX {
+  my ($dbh, $aref, @arg) = @_;
+
+  my ($cmd, %data);
+
+  foreach my $lref (@$aref) {	# loop over all retrieved data
+    next if $lref->[0] eq "YYYY/MM/DD";
+    $data{date} = UnixDate(ParseDate($lref->[0]), "%Y%m%d");
+    my $i = 1;
+    foreach my $fx (@arg) {
+      if (ExistsFXDailyData($dbh,$fx,%data)) { # update data if it exists
+	$cmd = "update fxprices set ";
+	$cmd .= "day_close = " . 1.0/$lref->[$i] . " "  .
+   	    "where currency = '$fx' and date  = '$data{date}'";
+      } else {
+	$cmd  = "insert into fxprices (currency, date, day_close) ";
+        $cmd .= "values ('$fx', '$data{date}', 1.0/$lref->[$i] )";
+      }
+      $i++;
+      if ($Config{commit}) {
+	print "$cmd\n" if $Config{debug};
+	$dbh->do($cmd) or die $dbh->errstr;
+      }
+    }
+    #print "Done with $fx (using $symbol)\n" if $Config{verbose};
+  }
+  if ($Config{commit}) {
+    $dbh->commit();
+  }
+}
 
 sub DatabaseInfoData {		# initialise a row in the info table
   my ($dbh, %hash) = @_;
@@ -1330,43 +1410,36 @@ sub DatabaseInfoData {		# initialise a row in the info table
 
 sub ExistsInfoSymbol {
   my ($dbh, %hash) = @_;
-  my $stmt = qq{select symbol
-                from stockinfo
-                where symbol = ?
-                };
-
-  print "ExistsInfoSymbol():\n\$stmt = $stmt\n" if $Config{debug};
-
-  if (@row = $dbh->selectrow_array($stmt, undef, $hash{symbol})) {
-    # plausibility tests here
-    # someone might care to extend this to consider the 'active' tuple
-    # maybe if it's false that fact should be noted since
-    # the user has apparently seen fit to add it to the database (again)
-    print "ExistsInfoSymbol:@row\n" if $Config{debug};
-    return 1;    # perl truth is > 0
-  } else {
-    return 0;
+  if (!defined($_symExistsInfoSymbolSth)) {
+      $_symExistsInfoSymbolSth = $dbh->prepare(qq{select symbol from stockinfo
+						  where symbol = ?})
+	  or die $dbh->errstr;
   }
+  $_symExistsInfoSymbolSth->execute($hash{symbol})
+      or die $_symExistsInfoSymbolSth->errstr;
+  my @rows = $_symExistsInfoSymbolSth->fetchrow_array();
+  $_symExistsInfoSymbolSth->finish();
+
+  # plausibility tests here
+  # someone might care to extend this to consider the 'active' tuple
+  # maybe if it's false that fact should be noted since
+  # the user has apparently seen fit to add it to the database (again)
+  return (@rows > 0);
 }
 
 
-sub ExistsDailyData {
+sub ExistsDailyData($%) {
   my ($dbh, %hash) = @_;
-  my $stmt = qq{select previous_close, day_open, day_low, day_high,
-		       day_close, day_change, bid, ask, volume
-		from stockprices
-		where symbol = ?
-		and date     = ?
-	       };
-
-  print "ExistsDailyData():\n\$stmt = $stmt\n" if $Config{debug};
-
-  if (@row = $dbh->selectrow_array($stmt, undef, $hash{symbol}, $hash{date})) {
-    # plausibility tests here
-    return 1;
-  } else {
-    return 0;
+  if (!defined($_symExistsDailyDataSth)) {
+      $_symExistsDailyDataSth = $dbh->prepare(qq{select symbol from stockprices
+					where symbol = ? and date = ?})
+	  or die $dbh->errstr;
   }
+  $_symExistsDailyDataSth->execute($hash{symbol}, $hash{date})
+      or die $_symExistsDailyDataSth->errstr;
+  my @rows = $_symExistsDailyDataSth->fetchrow_array();
+  $_symExistsDailyDataSth->finish();
+  return (@rows > 0);
 }
 
 
@@ -1383,12 +1456,9 @@ sub ExistsFXDailyData {
 
   my $sth = $dbh->prepare($stmt);
   $sth->execute($fx,$hash{date});
-  if (@row = $sth->fetchrow_array()) {
-    # plausibility tests here
-    return 1;
-  } else {
-    return 0;
-  }
+  my @row = $sth->fetchrow_array();
+  $sth->finish();
+  return (@row > 0);
 }
 
 
@@ -1434,6 +1504,9 @@ sub ParseDailyData {		# stuff the output into the hash
     if ($ra->[20] =~ m/(\S*)B$/) {
       # convert to millions from billions
       $hash{$key}{market_capitalisation} = $1*(1e3);
+    } elsif ($ra->[20] =~ m/(\S*)T$/) {
+      # reported in trillions -- convert to millions
+      $hash{$key}{market_capitalisation} = $1*(1e6);
     } elsif ($ra->[20] =~ m/(\S*)M$/) {
       # keep it in millions
       $hash{$key}{market_capitalisation} = $1;
@@ -1547,8 +1620,7 @@ sub ScrubDailyData {          # stuff the output into the hash
   foreach my $key (keys %hash) {# now check the date
     if ($hash{$key}{date} eq "N/A") { # if Yahoo! gave us no data
       if ($hash{$key}{symbol} =~ /^\^X/) { # and it was currency
-	my ($iso2yahoo,$yahoo2iso) = GetFXMaps;
-	my $retry = $yahoo2iso->{$hash{$key}{symbol}} . "USD=X";
+	my $retry = GetIsoCurrency($hash{$key}{symbol}) . "USD=X";
 	my @retrysymbols;
 	push @retrysymbols, $retry;	
 	my (@newarr) = GetDailyData(@retrysymbols);
@@ -1574,7 +1646,7 @@ sub ScrubDailyData {          # stuff the output into the hash
     if ($hash{$key}{date} ne $Config{today}) {   # if date is not today
 
       my $age = Delta_Format(DateCalc($hash{$key}{date}, $Config{lastbizday},
-				      undef, 2), 0, "%dt");
+				      undef, 2), "approx", 0, "%dt");
       if ($age > 5) {
         warn "Ignoring $hash{$key}{symbol} ($hash{$key}{name}) " .
 	  "with old date $hash{$key}{date}\n";
@@ -1666,8 +1738,6 @@ sub UpdateDatabase {		# update content in the db at end of day
 sub UpdateFXDatabase {
   my ($dbh, $res) = @_;
 
-  my ($iso2yahoo,$yahoo2iso) = GetFXMaps;
-
   # get all non-USD symbols (no USD as we don't need a USD/USD rate)
   my $stmt = qq{  select distinct currency
 		  from portfolio 
@@ -1678,12 +1748,12 @@ sub UpdateFXDatabase {
 
   print "UpdateFXDatabase():\n\$stmt = $stmt\n" if $Config{debug};
 
-  my @symbols = map { $iso2yahoo->{$ARG} } @{ $dbh->selectcol_arrayref($stmt)};
+  my @symbols = map { GetYahooCurrency($ARG) } @{ $dbh->selectcol_arrayref($stmt)};
   print "UpdateFXDatabase(): Symbols are ", join(" ", @symbols), "\n" 
       if $Config{debug};
   if ($Config{extrafx}) {
     foreach my $arg (split /,/, $Config{extrafx}) {
-      push @symbols, $iso2yahoo->{$arg};	
+      push @symbols, GetYahooCurrency($arg);	
     }
   }
   if (scalar(@symbols) > 0) {	# if there are FX symbols
@@ -1697,7 +1767,7 @@ sub UpdateFXDatabase {
 }
 
 ## use alternate FX data supply from the PACIFIC / Sauder School / UBC
-sub UpdateFXviaPACIFIC {
+sub UpdateFXviaUBC {
   my ($dbh, $res) = @_;
 
   # get all non-USD symbols (no USD as we don't need a USD/USD rate)
@@ -1707,10 +1777,10 @@ sub UpdateFXviaPACIFIC {
 		  and currency != 'USD'
 	    };
   $stmt .= "   and $res " if (defined($res));
-  print "UpdateFXviaPACIFIC():\n\$stmt = $stmt\n" if $Config{debug};
+  print "UpdateFXviaUBC():\n\$stmt = $stmt\n" if $Config{debug};
 
   my @symbols = @{ $dbh->selectcol_arrayref($stmt) };
-  print "UpdateFXviaPACIFIC() -- symbols=" . 
+  print "UpdateFXviaUBC() -- symbols=" . 
       join(" ", @symbols) . "\n" if $Config{debug};
 
   my %data;
@@ -1719,7 +1789,7 @@ sub UpdateFXviaPACIFIC {
 
   ## also fetch data via the PACIFIC server at Sauder / UBC
   my $ubcfx = GetUBCFXHash(\@symbols, $data{date}, $data{date});
-  print "PACIFIC server results\n", Dumper($ubcfx) if $Config{debug};
+  print "UBC server results\n", Dumper($ubcfx) if $Config{debug};
 
   foreach my $key (keys %{$ubcfx}) { # split these into reference to the arrays
     my $fx = $key; #$yahoo2iso->{$hash{$key}{symbol}};
@@ -1751,6 +1821,7 @@ sub UpdateFXviaPACIFIC {
 	my $sth = $dbh->prepare($stmt);
 	$sth->execute($fx, $data{date}, 1/$ubcfx->{$fx})
 	  or warn "Failed for $fx at $data{date}\n";
+	$sth->finish();
       }
     }
     if ($Config{commit}) {
@@ -1934,7 +2005,7 @@ F<LWP.3pm>, F<Date::Manip.3pm>
 
 =head1 COPYRIGHT
 
-Finance::BeanCounter.pm  (c) 2000 -- 2005 by Dirk Eddelbuettel <edd@debian.org>
+Finance::BeanCounter.pm  (c) 2000 -- 2006 by Dirk Eddelbuettel <edd@debian.org>
 
 Updates to this program might appear at 
 F<http://eddelbuettel.com/dirk/code/beancounter.html>.
